# coding=utf-8
"""
Handles the station overview page and any other station-level pages.
"""

from datetime import datetime, timedelta
import web
from web.contrib.template import render_jinja
from cache import day_cache_control
import config
from database import get_daily_records, get_years, total_rainfall_in_last_7_days, day_exists
import os
from months import month_name
from ui import get_nav_urls
from ui import validate_request, html_file

__author__ = 'David Goodwin'

basic_template_dir = os.path.join(os.path.dirname(__file__),
                                  os.path.join('basic_templates'))
modern_template_dir = os.path.join(os.path.dirname(__file__),
                                   os.path.join('modern_templates'))

basic_templates = render_jinja(basic_template_dir, encoding='utf-8')
modern_templates = render_jinja(modern_template_dir, encoding='utf-8')

def get_station_standard(ui, station):
    """
    Index page for the weather station. Should give station records and
    perhaps some basic yearly overview data.
    :param ui: The UI being used
    :type ui: str
    :param station: Name of the station to show info for.
    :return: View data.
    """

    current_location = '/s/' + station + '/'

    now = datetime.now().date()

    class data:
        """ Data required by the view """
        records = get_daily_records(now)
        years = get_years()
        year = now.year
        month_s = month_name[now.month]
        yesterday = now - timedelta(1)
        yesterday_month_s = month_name[yesterday.month]
        rainfall_7days_total = total_rainfall_in_last_7_days(now)

    if not day_exists(data.yesterday):
        data.yesterday = None
        data.yesterday_month_s = None

    if data.records is None:
        return web.NotFound(message="No Data")

    if ui == 'm':
        sub_dir = ''
    else:
        sub_dir = 'datatable/'

    day_cache_control(None, now)
    nav_urls = get_nav_urls(station, current_location)
    return modern_templates.station(nav=nav_urls,
                                    data=data,
                                    station=station,
                                    ui=ui,
                                    sitename=config.site_name,
                                    subdir=sub_dir)

def get_station_basic(station):
    """
    Index page for the weather station. Should give station records and
    perhaps some basic yearly overview data.
    :param station:
    :return:
    """

    return basic_templates.station(years=get_years(),station=station)

class station:
    """
    Index page for the station. Should give station records, basic yearly
    overview data, etc.
    """
    def GET(self, ui, station):
        """
        Station overview page (on the standard UI) or year list page (basic UI)
        :param ui: UI to use
        :type ui: str
        :param station: Station to get data for
        :type station: str
        :return: HTML data.
        """
        validate_request(ui,station)
        html_file()

        if ui in ('s','m'):
            return get_station_standard(ui, station)
        else:
            return get_station_basic(station)